/**
 * 
 */
package org.processmining.models.operationalsupport;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.HashSet;

import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XTrace;
import org.deckfour.xes.model.impl.XAttributeMapImpl;
import org.deckfour.xes.model.impl.XTraceImpl;
import org.processmining.models.operationalsupport.xml.OSXMLConverter;

/**
 * Data structure for storing a request requesting recommendation for scheduling
 * in a process management system.
 * 
 * @author Christian W. Guenther (christian@deckfour.org)
 * 
 */
public class Request {

	// Host ID management
	protected static String hostId = "UNKNOWN_HOST";
	static {
		try {
			InetAddress addr = InetAddress.getLocalHost();
			hostId = addr.getCanonicalHostName() + "(" + addr.getHostAddress() + ")";
		} catch (UnknownHostException e) {
		}
	}

	protected String id;
	protected String processId;
	protected String processInstanceId;
	protected HashSet<WorkItem> availableItems;
	protected XTrace trace;

	/**
	 * Creates an empty recommendation request with a known ID
	 */
	public Request(String requestId, String aProcessId, String aProcessInstanceId) {
		id = requestId;
		processId = aProcessId;
		processInstanceId = aProcessInstanceId;
		availableItems = new HashSet<WorkItem>();
		trace = new XTraceImpl(new XAttributeMapImpl());
	}

	/**
	 * Creates a new, empty recommendation request
	 */
	public Request(String aProcessId, String aProcessInstanceId) {
		this(Request.hostId + "_" + Long.toHexString(System.currentTimeMillis()), aProcessId, aProcessInstanceId);
	}

	/**
	 * Resets the request id.
	 * 
	 * @param id
	 *            representsd the new id
	 */
	public void resetId(String id) {
		this.id = id;
	}

	/**
	 * Adds a user to the set of available resources to be used for
	 * recommendation.
	 * 
	 * @param user
	 *            string representation of a user
	 */
	public boolean addAvailableItem(WorkItem item) {
		return availableItems.add(item);
	}

	/**
	 * Adds an audit trail entry, or event, to the current incomplete audit
	 * trail of the process instance for which the request is being issued.
	 * 
	 * @param ate
	 *            an audit trail entry (event).
	 */
	public void addAllEvents(XTrace t) {
		trace.addAll(t);
	}

	/**
	 * Adds an audit trail entry, or event, to the current incomplete audit
	 * trail of the process instance for which the request is being issued.
	 * 
	 * @param ate
	 *            an audit trail entry (event).
	 */
	public void addEvent(XEvent ate) {
		trace.add(ate);
	}

	/**
	 * @return A string containing the unique ID of this request (used for
	 *         correlation)
	 */
	public String getId() {
		return id;
	}

	/**
	 * @return the unique ID of the process this request refers to
	 */
	public String getProcessId() {
		return processId;
	}

	/**
	 * @return the unique ID of the process instance this request refers to
	 */
	public String getProcessInstanceId() {
		return processInstanceId;
	}

	/**
	 * @return A sorted set of strings specifying the users currently available
	 *         for scheduling
	 */
	public Iterable<WorkItem> getAvailableItems() {
		return availableItems;
	}

	/**
	 * @return The current, incomplete, execution trace (i.e. sequence of
	 *         events) of the process instance for which the request is issued
	 */
	public XTrace getTrace() {
		return trace;
	}

	public String toString() {
		try {
			return (new OSXMLConverter()).requestToXML(this);
		} catch (Exception e) {
			e.printStackTrace();
			return "ERROR";
		}
	}

}
