package org.processmining.models.operationalsupport.net.service;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.concurrent.Executor;

import org.processmining.framework.connections.ConnectionManager;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.PluginManager;
import org.processmining.framework.plugin.ProMFuture;
import org.processmining.framework.plugin.events.ProvidedObjectLifeCycleListener;
import org.processmining.framework.providedobjects.ContextAwareObject;
import org.processmining.framework.providedobjects.ProvidedObjectDeletedException;
import org.processmining.framework.providedobjects.ProvidedObjectID;
import org.processmining.framework.providedobjects.ProvidedObjectManager;
import org.processmining.framework.util.socket.Service;
import org.processmining.framework.util.socket.ServiceEnvironment;
import org.processmining.models.operationalsupport.ProviderResponse;
import org.processmining.models.operationalsupport.Request;
import org.processmining.models.operationalsupport.Response;
import org.processmining.models.operationalsupport.net.provider.Provider;
import org.processmining.models.operationalsupport.net.provider.impl.AbstractProvider;

public class OSService extends OSServiceHandler implements Serializable, ProvidedObjectLifeCycleListener,
		ContextAwareObject {

	private static final long serialVersionUID = -8673517983880516654L;

	/**
	 * The service is defined transient, as it should not be serialized. Upon
	 * deserialization, a new service is constructed on the original port, but
	 * this port is not opened. Instead, a new manager should be started on the
	 * service.
	 */
	private transient Service requestService;

	private transient ServiceEnvironment environment;

	/**
	 * The service's running state is defined transient. Upon deserialization,
	 * this flag is set to false. A new manager should be started on the service
	 * to set it to true again.
	 */
	private transient boolean running = false;

	private transient ProvidedObjectID poID = null;

	private transient ProvidedObjectManager manager;

	private transient Map<ProvidedObjectID, Provider> providers;
	private transient Map<ProvidedObjectID, ProMFuture<? extends Provider>> futureProviders;

	private final int requestPort;

	public OSService(int requestPort, int portOffset, int maxClients, ProvidedObjectManager manager) {
		super(portOffset, maxClients);
		this.requestPort = requestPort;
		setProvidedObjectManager(manager);
		setupService();
	}

	/*
	 * public int getrequestPort(){ return requestPort; }
	 * 
	 * public int getInfoPort(){ return infoPort; }
	 */

	public String toString() {
		return Integer.toString(requestPort);
	}

	private void setupService() {
		requestService = new Service(requestPort);
		environment = null;
		running = false;
	}

	public void stop(ServiceEnvironment environment) {
		assert (environment.isCancelled());
		setupService();
	}

	public void start(ServiceEnvironment environment, Executor executor) throws IOException {
		this.environment = environment;
		boolean start = false;
		synchronized (this) {
			if (!running) {
				running = true;
				start = true;
			}
		}
		if (start) {
			requestService.addServiceHandler(this);
			requestService.start(environment, executor);
		}

	}

	protected Response getResult(ServiceEnvironment environment, Request request) {
		Response result = new Response(request.getId()); // create a new OS response for the request
		HashMap<ProvidedObjectID, Provider> localProviders = new HashMap<ProvidedObjectID, Provider>(providers);
		for (Map.Entry<ProvidedObjectID, Provider> provider : localProviders.entrySet()) { // loop through all registered providers
			// Get the label of the provider as set by the user in the UI:
			String name;
			try {
				name = manager.getProvidedObjectLabel(provider.getKey());
			} catch (ProvidedObjectDeletedException e) {
				// Removal will be handled later, just skip this provider for now.
				continue;
			}
			ProviderResponse response = new ProviderResponse(name);
			provider.getValue().populateResponse(environment, request, response);
			result.add(response); // add the provider response to the OS response
			System.out.println();
		}

		return result;
	}

	private void writeObject(ObjectOutputStream out) throws IOException {
		out.defaultWriteObject();
	}

	private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {
		in.defaultReadObject();
		setupService();
		setupPorts();
	}

	public void providedObjectCreated(ProvidedObjectID objectID, PluginContext context) {
		providedObjectCreated(objectID);
	}

	@SuppressWarnings("unchecked")
	private void providedObjectCreated(ProvidedObjectID objectID) {
		try {
			Class<?> type = manager.getProvidedObjectType(objectID);
			Object object = manager.getProvidedObjectObject(objectID, false);
			if (AbstractProvider.class.isAssignableFrom(type)) {
				if (object instanceof ProMFuture) {
					//System.out.println("Added a future provider to " + this.hashCode());
					futureProviders.put(objectID, (ProMFuture<? extends AbstractProvider>) object);
				} else {
					AbstractProvider provider = (AbstractProvider) object;
					if (provider.getOwner() == this) {
						synchronized (providers) {
							//System.out.println("Added a provider to " + this.hashCode());
							providers.put(objectID, provider);
						}
					}
				}
			}
		} catch (ProvidedObjectDeletedException e) {
			// ignore
		}
	}

	public void providedObjectDeleted(ProvidedObjectID objectID) {
		synchronized (providers) {
			providers.remove(objectID);
			futureProviders.remove(objectID);
		}
		if (objectID.equals(poID)) {
			// The service was deleted...
			HashSet<ProvidedObjectID> tempProviders = new HashSet<ProvidedObjectID>(providers.keySet());
			// delete all providers
			for (ProvidedObjectID id : tempProviders) {
				try {
					manager.deleteProvidedObject(id);
				} catch (ProvidedObjectDeletedException e) {
					// gracefully ignore.
				}
			}
			if (environment != null) {
				environment.cancel();
			}
		}
	}

	public void providedObjectFutureReady(ProvidedObjectID objectID) {
		synchronized (futureProviders) {
			if (futureProviders.containsKey(objectID)) {
				ProMFuture<? extends Provider> future = futureProviders.get(objectID);
				try {
					Provider provider = future.get();
					if (provider.getOwner().equals(this)) {
						synchronized (providers) {
							//							System.out.println("Added a finished provider to " + this.hashCode());
							providers.put(objectID, provider);
						}
					}
				} catch (Exception e) {
					// ignore then
				} finally {
					futureProviders.remove(objectID);
					//					System.out.println("Removed a future provider from " + this.hashCode());
				}
			}
		}
		Object o;
		try {
			o = manager.getProvidedObjectObject(objectID, false);
			if (o == this) {
				poID = objectID;
			}
		} catch (ProvidedObjectDeletedException e) {
			// gracefully ignore
		}
	}

	public void providedObjectNameChanged(ProvidedObjectID objectID) {
		// Don't care
	}

	public void providedObjectObjectChanged(ProvidedObjectID objectID) {
		// Don't care
	}

	protected String processInfo(ServiceEnvironment environment, String provider, String info) {
		return null;
	}

	public void setManagers(ConnectionManager connectionManager, PluginManager pluginManager,
			ProvidedObjectManager providedObjectManager) {
		setProvidedObjectManager(providedObjectManager);
	}

	private void setProvidedObjectManager(ProvidedObjectManager manager) {

		// set the manager
		this.manager = manager;
		providers = Collections.synchronizedMap(new HashMap<ProvidedObjectID, Provider>());

		futureProviders = Collections.synchronizedMap(new HashMap<ProvidedObjectID, ProMFuture<? extends Provider>>());

		for (ProvidedObjectID id : manager.getProvidedObjects()) {
			providedObjectCreated(id);
		}

		// register as a listener.
		this.manager.getProvidedObjectLifeCylceListeners().add(this);

	}

}